-- See the GranSim user's guide for a description

module Main where

import System
import Strategies
import qualified Parallel

{-

Although the sum-of-squares program is a very simple producer-consumer
program, it already shows some basic problems in this model.
Unfortunately, the simple version of the program, which was presented
in section Semi-explicit Parallelism, produces hardly any parallelism
at all.  The reason for this behaviour is that because of the lazy
evaluation approach the producer will only create the top cons cell of
the intermediate list.  When the consumer then demands the rest of the
list it computes squares as well as the result sum.  In the example
below version 1 (with res_naive) shows this behaviour.

To improve the parallelism one can use a forcing function on the
intermediate list.  The parallel thread is then the application of
this forcing function on the squares list.  As expected, this creates
two threads that are computing most of the time and occasionally have
to exchange data.  Version 2 (with res_force) shows this behaviour.

A better way to achieve the same operational behaviour is to define a
strategy, how to compute the result.  The strategy describes the
parallelism and the evaluation degree.  In doing so, it uses the
reduce to normal form (rnf) strategy, which is defined in the class
NFData.  Compared to version 2 the strategy version 3 (with
res_strategy) achieves a clean separation of defining the result and
specifying operational details like parallelism and evaluation degree.

-}

main = getArgs >>= \[arg1, arg2] ->
       let 
	 version = read arg1
	 n = read arg2

         squares :: [Int] 
         squares = [ i^2 | i <- [1..n] ]

         -- Naive version sparks a producer for list squares but
         -- doesn't force it

         res_naive = squares `Parallel.par` sum squares

         -- This version sparks a forcing function on the list (producer)

         res_force = (foldl1 `Parallel.seq` squares) `Parallel.par` sum squares

         -- The strategy version

         res_strat = sum squares `using` (par (rnf squares) . rnf)

         res = case version of
                1 -> res_naive               
                2 -> res_force               
                3 -> res_strat               
                _ -> res_naive
         str = case version of
                1 -> "Naive version"
                2 -> "Forcing version"               
                3 -> "Strategy version"
                _ -> "Naive version"
       in 
       putStr ("Sum of squares (" ++ str ++ ") of length " ++ (show n) ++ 
              " = " ++ (show res) ++ "\n")

